/*
 * Copyright (c) 2013 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */
package org.processmining.xeslite.external;

import java.io.IOException;
import java.util.ArrayDeque;
import java.util.Collections;
import java.util.Comparator;
import java.util.Deque;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentNavigableMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedTransferQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.locks.LockSupport;
import java.util.prefs.Preferences;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.mapdb.BTreeKeySerializer;
import org.mapdb.DB;
import org.mapdb.Fun;
import org.mapdb.Fun.Pair;

import com.google.common.collect.Interner;
import com.google.common.collect.Interners;
import com.google.common.collect.Iterators;
import com.google.common.primitives.Ints;

/**
 * Implementation using MapDB (https://github.com/jankotek/mapdb) as back-end.
 * 
 * @author F. Mannhardt
 * 
 */
public final class MapDBStore implements ExternalStore {
	
	private static final Preferences PREFS = Preferences.userRoot().node("org.processmining.xeslite.external");
	
	private static final int LITERAL_POOL_SIZE = 1 << 17;
	public static final int NODE_SIZE = 32;
	
	final class PumpServiceImpl implements PumpService {

		private final class PumpIterator implements Iterator<Fun.Pair<Long, ExternalAttribute>> {
			
			private final class AttributeComparator implements Comparator<XAttribute> {
				public int compare(XAttribute a1, XAttribute a2) {
					int key1, key2 = 0;
					if (a1 instanceof ExternalAttribute) {
						key1 = ((ExternalAttribute) a1).getInternalKey();
					} else {
						assert a1.getKey() != null : "Key null for "+a1;
						key1 = keyPool.put(a1.getKey());
					}
					if (a2 instanceof ExternalAttribute) {
						key2 = ((ExternalAttribute) a2).getInternalKey();
					} else {
						assert a2.getKey() != null : "Key null for "+a2;
						key2 = keyPool.put(a2.getKey());
					}
					return Ints.compare(key2, key1); // Reverse order
				}
			}

			private final Deque<Pair<XAttributable, List<XAttribute>>> buffer = new ArrayDeque<Fun.Pair<XAttributable,List<XAttribute>>>();
			private ExternalAttributable owner;
			private Iterator<XAttribute> attributes = Iterators.emptyIterator();
			private Comparator<? super XAttribute> comparator = new AttributeComparator();

			public Pair<Long, ExternalAttribute> next() {
				assert owner != null;
				assert owner.getExternalId() != Long.MIN_VALUE;
				ExternalAttribute a = XAttributeExternalImpl.convert(MapDBStore.this, attributes.next());
				if (owner instanceof AttributesCacheable) {
					AttributesCacheable cacheable = (AttributesCacheable) owner;
					switch (a.getKey()) {
						case XConceptExtension.KEY_NAME :
							cacheable.setName(interner.intern(((XAttributeLiteral) a).getValue()));
							break;
						case XTimeExtension.KEY_TIMESTAMP :
							cacheable.setTime(((XAttributeTimestamp) a).getValueMillis());
							break;

						case XLifecycleExtension.KEY_TRANSITION :
							cacheable.setTransition(interner.intern(((XAttributeLiteral) a).getValue()));
							break;

						default :
							break;
					}
				}
				Long compositeKey = getCompositeKey(owner, a.getInternalKey());
				return new Pair<>(compositeKey, a);
			}

			public boolean hasNext() {
				try {
					if (attributes == null) {
						assert owner == null;
						return false;
					}
					if (attributes.hasNext()) {
						assert owner != null;
						return true;
					}
					Pair<XAttributable, List<XAttribute>> attributesWithOwner = takeNext();
					if (attributesWithOwner != endOfPump) {
						sortDecreasingOrder(attributesWithOwner.b);
						owner = (ExternalAttributable) attributesWithOwner.a;
						attributes = attributesWithOwner.b.iterator();
						if (!attributes.hasNext()) {
							throw new RuntimeException("Empty list of attributes is not allowed!");
						}
						return true;
					} else {
						return endOfData();
					}
				} catch (InterruptedException e) {
					return endOfData();
				}
			}

			private final void tryFillBuffer() throws InterruptedException {
				if (buffer.isEmpty()) {
					if (pumpQueue.drainTo(buffer) == 0) {
						// Add at least one
						buffer.add(pumpQueue.take());
					}
				}								
			}

			private final Pair<XAttributable, List<XAttribute>> takeNext() throws InterruptedException {
				tryFillBuffer();
				return buffer.poll();
			}

			private final boolean endOfData() {
				owner = null;
				attributes = null;
				return false;
			}

			private final void sortDecreasingOrder(List<XAttribute> attributes) {
				// Sort attributes in decreasing order for pump
				Collections.sort(attributes, comparator);
			}

			public void remove() {
				throw new UnsupportedOperationException();
			}
		}

		private final ExecutorService pumpExecutor = Executors.newSingleThreadExecutor();
		private final Future<ConcurrentNavigableMap<Long, ExternalAttribute>> pumpFuture;

		private final BlockingQueue<Pair<XAttributable, List<XAttribute>>> pumpQueue = new LinkedTransferQueue<>();
		private final Iterator<Pair<Long, ExternalAttribute>> pumpIterator;		
		
		private final Pair<XAttributable, List<XAttribute>> endOfPump = new Pair<>(null, null);
		
		/**
		 * Starts the pump thread. Side-effect: The 'db' will be locked!
		 */
		public PumpServiceImpl() {
			db.delete(".attributeStore");
			assert idFactory instanceof IdFactorySeqDecreasing : "Keys need to be sorted in decreasing order for data pump";
			pumpIterator = new PumpIterator();
			final Thread callingThread = Thread.currentThread(); // Use to interrupt calling thread upon error on pump thread
			// Auto-start the creation thread - be aware that 'db' is locked now!
			// So it cannot be used until finish pump is called!
			pumpFuture = pumpExecutor.submit(new Callable<ConcurrentNavigableMap<Long, ExternalAttribute>>() {

				public ConcurrentNavigableMap<Long, ExternalAttribute> call() throws Exception {
					try {
						return db.treeMapCreate(".attributeStore").keySerializer(BTreeKeySerializer.LONG)
								.valueSerializer(new MapDBAttributeSerializer()).nodeSize(NODE_SIZE).pumpSource(pumpIterator)
								.makeOrGet();
					} catch (Exception e) {
						callingThread.interrupt();
						throw e;
					}
				}
			});
			
		}
		
		public void pumpAttributes(XAttributable attributable, List<XAttribute> attributes) {
			if (attributable instanceof ExternalAttributable) {
					if (!attributes.isEmpty()) {
						boolean result = pumpQueue.offer(new Pair<XAttributable, List<XAttribute>>(attributable, attributes));
						while (!result) {
							// Back-off wait until consumer keeps up
							LockSupport.parkNanos(100l);
							result = pumpQueue.offer(new Pair<XAttributable, List<XAttribute>>(attributable, attributes));
						}
					}
			} else {
				throw new RuntimeException("Cannot pump " + attributable + " it does not implement "
						+ ExternalAttributable.class.getSimpleName());
			}
		}

		public void finishPump() {
			try {
				pumpQueue.offer(endOfPump);
				try {
					attributeStorage = pumpFuture.get();
				} catch (ExecutionException e) {
					throw new RuntimeException(e);
				}
			} catch (InterruptedException e) {
				// Catch to find out the real reason for this
				try {
					pumpFuture.get(1, TimeUnit.SECONDS);
					// There was no exception, just interrupt again
					Thread.currentThread().interrupt();
				} catch (ExecutionException | TimeoutException | InterruptedException e1) {
					throw new RuntimeException(e1);
				}
			} finally {
				pumpExecutor.shutdown();
				MapDBStore.this.pumpService = null; // Remove reference to ourselves for GC
			}
		}

	}

	public static final class Builder {

		private MapDBDatabase database;
		private boolean isPump = false;

		public Builder withDatabase(MapDBDatabase database) {
			this.database = database;
			return this;
		}
		
		public Builder withPump() {
			this.isPump = true;
			return this;
		}

		public MapDBStore build() {
			if (database == null)
				// Stand-alone mode with temporary file
				database = new MapDBDatabaseImpl();
			try {
				database.createDB();
			} catch (IOException e) {
				throw new RuntimeException("Failed to access MapDB database file!");
			}
			return new MapDBStore(this);
		}

	}

	private PumpService pumpService;
	
	private ConcurrentNavigableMap<Long, ExternalAttribute> attributeStorage;
	private IdFactory idFactory;

	private final DB db;
	private final StringPool keyPool;
	private final StringPool literalPool;
	private final Interner<String> interner;

	private boolean useCachingWrapper = true;

	private MapDBStore(Builder builder) {
		super();
		this.db = builder.database.getDB();
		this.interner = Interners.newStrongInterner();
		
		if (builder.isPump) {			
			this.idFactory = new IdFactorySeqDecreasing(getKeyPoolShift()); // decreasing ids for data pump		
			this.keyPool = createKeyPool(db, idFactory.getIntervalCapacity());
			this.literalPool = createLiteralPool(db, LITERAL_POOL_SIZE);	  
		} else {
			this.idFactory = createIDFactory(db);
			this.keyPool = createKeyPool(db, idFactory.getIntervalCapacity());
			this.literalPool = createLiteralPool(db, LITERAL_POOL_SIZE);
			if (db.exists(".attributeStore")) {
				attributeStorage = db.treeMap(".attributeStore");
			} else {
				attributeStorage = db.treeMapCreate(".attributeStore").keySerializer(BTreeKeySerializer.LONG)
						.valueSerializer(new MapDBAttributeSerializer())
						// Trading random access speed for faster sequential access
						// .valuesOutsideNodesEnable()					
						.nodeSize(NODE_SIZE) // 32 is a good compromise for concurrent writes
						.makeOrGet();
			}
		}
		
	}

	private static StringPool createKeyPool(DB db, int intervalCapacity) {
		StringPool pool = db.<StringPool>get(".stringPool");
		if (pool != null) {
			return pool;
		} else {
			return new StringPoolCASImpl(intervalCapacity);
		}
	}

	private static StringPool createLiteralPool(DB db, int intervalCapacity) {
		StringPool pool = db.<StringPool>get(".literalPool");
		if (pool != null) {
			return pool;
		} else {
			return new StringPoolCASImpl(intervalCapacity * 2);
		}
	}

	private static IdFactory createIDFactory(DB db) {
		return new IdFactoryRandomBucket(getKeyPoolShift(), db.<Long>get(".seqNr"));
	}

	private static int getKeyPoolShift() {
		return PREFS.getInt("keyPoolShift", 14);
	}

	/* (non-Javadoc)
	 * @see org.processmining.xeslite.external.ExternalStore#startPump()
	 */
	public PumpService startPump() {
		pumpService = new PumpServiceImpl();
		return pumpService;
	}

	public PumpService getPumpService() {
		return pumpService;
	}

	public boolean isPumping() {
		return pumpService != null;
	}

	final ConcurrentNavigableMap<Long, ExternalAttribute> getMapStorage() {
		return attributeStorage;
	}

	/**
	 * Returns the composite key of the attribute and the owner. Assumes that
	 * the internal key of the attribute is already added to the pool.
	 * 
	 * @param owner
	 * @param attributekey
	 * @return the composite key
	 */
	static Long getCompositeKey(ExternalAttributable owner, int attributekey) {
		return owner.getExternalId() + attributekey;
	}

	/**
	 * @param owner
	 * @param compositeKey
	 * @return
	 */
	static int getInternalAttributeKey(ExternalAttributable owner, long compositeKey) {
		return (int) (compositeKey - owner.getExternalId());
	}

	/**
	 * Returns the internal key of the attribute with the supplied
	 * {@link String} key. In case no such attribute is know, NULL is returned.
	 * 
	 * @param owner
	 * @param attributeKey
	 * @return the composite key or NULL is no such key exists
	 */
	final Integer getInternalAttributeKey(ExternalAttributable owner, String attributeKey) {
		return keyPool.getIndex(attributeKey);
	}

	/**
	 * @param owner
	 * @return
	 */
	final long lowestCompositeKey(ExternalAttributable owner) {
		return owner.getExternalId();
	}

	/**
	 * @param owner
	 * @return
	 */
	final long highestCompositeKey(ExternalAttributable owner) {
		return owner.getExternalId() + (getAttributeKeyPool().getPoolLimit() - 1);
	}

	private final XAttributeMap getAttributeMap(ExternalAttributable attributable) {
		XAttributeMap attributeMap = null;
		if (useCachingWrapper) {
			attributeMap = new MapDBAttributeMapWithCache(attributable, this, interner);
		} else {
			attributeMap = new MapDBAttributeMap(attributable, this);
		}
		return attributeMap;
	}

	public final StringPool getAttributeKeyPool() {
		return keyPool;
	}

	public StringPool getLiteralPool() {
		return literalPool;
	}

	public final IdFactory getIdFactory() {
		return idFactory;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#getAttributes
	 * (long)
	 */
	@Override
	public final XAttributeMap getAttributes(final ExternalAttributable attributable) {
		if (isPumping()) {
			throw new IllegalStateException("Cannot be used during data pump!");
		}
		return getAttributeMap(attributable);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#hasAttributes
	 * (org.progressmining.xeslite.external.store.XExternalAttributable)
	 */
	@Override
	public final boolean hasAttributes(final ExternalAttributable attributable) {
		if (isPumping()) {
			throw new IllegalStateException("Cannot be used during data pump!");
		}
		//TODO somehow cache this
		return !attributeStorage.keySet()
				.subSet(lowestCompositeKey(attributable), true, highestCompositeKey(attributable), true).isEmpty();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#setAttributes
	 * (long, org.deckfour.xes.model.XAttributeMap)
	 */
	@Override
	public final void setAttributes(final ExternalAttributable attributable, final XAttributeMap attributes) {
		if (isPumping()) {
			throw new IllegalStateException("Cannot be used during data pump!");
		}
		
		if (attributes.isEmpty()) {
			getAttributeMap(attributable).clear();
			return;
		}

		XAttributeMap attributeMap = getAttributeMap(attributable);
		Iterator<XAttribute> values = attributeMap.values().iterator();
		while (values.hasNext()) {
			XAttribute a = values.next();
			if (!attributes.containsKey(a.getKey())) {
				values.remove();
			}
		}
		attributeMap.putAll(attributes);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.progressmining.xeslite.external.store.XAttributeStore#removeAttributes
	 * (long)
	 */
	@Override
	public final XAttributeMap removeAttributes(final ExternalAttributable attributable) {
		if (isPumping()) {
			throw new IllegalStateException("Cannot be used during data pump!");
		}
		
		XAttributeMap attributeMap = getAttributeMap(attributable);

		XAttributeMap oldMap = new XAttributeMapImpl(attributeMap.size());
		// Use values() instead of putAll() to avoid two copies (1 by entrySet() that is called by putAll(), 2 by us)
		Iterator<XAttribute> values = attributeMap.values().iterator();
		while (values.hasNext()) {
			XAttribute a = values.next();
			oldMap.put(a.getKey(), a);
		}
		attributeMap.clear();
		return oldMap;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.progressmining.xeslite.external.store.XAttributeStore#commit()
	 */
	@Override
	public final void commit() {
		if (isPumping()) {
			throw new IllegalStateException("Cannot be used during data pump!");
		}
		
		// Only does a sync
		db.commit();
	}

	/* (non-Javadoc)
	 * @see org.processmining.xeslite.external.ExternalStore#dispose()
	 */
	public void dispose() {
		if (isPumping()) {
			throw new IllegalStateException("Cannot be used during data pump!");
		}
		
		db.close();
	}

}